from typing import Union, Optional
import pathlib
import shutil
from numpy import dtype as np_dtype
from numpy import ndarray
from ctypes import c_size_t
from expelliarmus.wizard.clib import (
    event_t,
    dat_cargo_t,
    evt2_cargo_t,
    evt3_cargo_t,
    events_cargo_t,
)
from expelliarmus.wizard.wizard_wrapper import (
    c_read_wrapper,
    c_cut_wrapper,
    c_read_chunk_wrapper,
)
from expelliarmus.utils import (
    check_chunk_size,
    check_file_encoding,
    check_encoding,
    check_input_file,
    check_output_file,
    check_buff_size,
    check_external_file,
    check_new_duration,
    _DEFAULT_DTYPE,
    _DEFAULT_BUFF_SIZE,
)


class Wizard:
    """
    Wizard allows to cast powerful spells, such as reading entire files to NumPy arrays or chunks of these, or cutting binary files to a certain duration.

    :param encoding: the encoding of the file, to be chosen among DAT, EVT2 and EVT3.
    :param fpath: the file to be read, either in chunks or fully.
    :param buff_size: the size of the buffer used to read the binary file.
    :param chunk_size: the chunk lenght when reading files in chunks.
    """

    def __init__(
        self,
        encoding: str,
        fpath: Optional[Union[str, pathlib.Path]] = None,
        chunk_size: Optional[int] = 8192,
        buff_size: Optional[int] = _DEFAULT_BUFF_SIZE,
    ) -> None:
        self._buff_size = check_buff_size(buff_size)
        self._encoding = check_encoding(encoding)
        if fpath is None:
            self._fpath = None
        else:
            self._fpath = check_input_file(fpath=fpath, encoding=self._encoding)
        self._chunk_size = check_chunk_size(chunk_size, encoding)
        self.cargo = self._get_cargo()
        return

    @property
    def encoding(self) -> str:
        """
        The encoding of the files handles by Wizard.

        :returns: the encoding.
        """
        return self._encoding

    @property
    def fpath(self) -> pathlib.Path:
        """
        Wizard input file path.

        :returns: the file path.
        """
        return self._fpath

    @property
    def buff_size(self) -> int:
        """
        The buffer size used by Wizard to read the binary files.

        :returns: the buffer size.
        """
        return self._buff_size

    @property
    def chunk_size(self) -> int:
        """
        The length of each chunk provided by Wizard when reading files in chunks.

        :returns: the chunk size.
        """
        return self._chunk_size

    @encoding.setter
    def encoding(self, value):
        raise AttributeError("ERROR: Denied setting of private attribute _encoding.")

    @fpath.setter
    def fpath(self, value):
        raise AttributeError("ERROR: Denied setting of private attribute _fpath.")

    @buff_size.setter
    def buff_size(self, value):
        raise AttributeError("ERROR: Denied setting of private attribute _buff_size.")

    @chunk_size.setter
    def chunk_size(self, value):
        raise AttributeError("ERROR: Denied setting of private attribute _chunk_size.")

    def _get_cargo(self) -> Union[dat_cargo_t, evt2_cargo_t, evt3_cargo_t]:
        events_info = events_cargo_t(
            c_size_t(self.chunk_size),
            1,
            0,
        )
        if self.encoding == "DAT":
            cargo = dat_cargo_t(events_info, 0, 0)
        elif self.encoding == "EVT2":
            cargo = evt2_cargo_t(events_info, 0, 0)
        elif self.encoding == "EVT3":
            event = event_t(0, 0, 0, 0)
            cargo = evt3_cargo_t(events_info, 0, 0, 0, 0, 0, event)
        return cargo

    def set_file(self, fpath: Union[str, pathlib.Path]) -> None:
        """
        Function that sets the input file.

        :param fpath: the path to the file.
        """
        self._fpath = check_input_file(fpath, encoding=self.encoding)
        return

    def set_chunk_size(self, chunk_size: int) -> None:
        """
        Function to sets the size of the chunks to be read.
        WARNING: due to the vectorized events in event files, the chunks can be longer that 'chunk_size' (at most 'chunk_size+12').

        :param fpath: path to the input file.
        :param chunk_size: size of the chunks ot be read.
        """
        self._chunk_size = check_chunk_size(chunk_size, self.encoding)
        self.cargo = self._get_cargo()
        return

    def set_encoding(self, encoding: Union[str, pathlib.Path]) -> None:
        """
        Sets the encoding proprierty of the class.

        :param encoding: the encoding of the file.
        """
        self._encoding = check_encoding(encoding)
        return

    def set_buff_size(self, buff_size: int) -> None:
        """
        Sets te buffer size used to read the binary file.

        :param buff_size: the buffer size specified.
        """
        self._buff_size = check_buff_size(buff_size)
        return

    def reset(self) -> None:
        """
        Function used to reset the generator, so that the file can be read from the beginning.
        """
        if not (self.cargo is None):
            del self.cargo
        self.cargo = self._get_cargo()
        return

    def cut(
        self,
        fpath_out: Union[str, pathlib.Path],
        new_duration: int,
        fpath_in: Optional[Union[str, pathlib.Path]] = None,
    ) -> int:
        """
        Cuts the duration of the recording contained in 'fpath_in' to 'new_duration' and saves the result to 'fpath_out'.

        :param fpath_in: path to input file.
        :param fpath_out: path to output file.
        :param new_duration: the desired duration, expressed in milliseconds.

        :returns: the number of events encoded in the output file.
        """
        fpath_in = check_external_file(fpath_in, self.fpath, self.encoding)
        fpath_out = check_output_file(fpath=fpath_out, encoding=self.encoding)
        new_duration = check_new_duration(new_duration)
        nevents = c_cut_wrapper(
            encoding=self.encoding,
            fpath_in=fpath_in,
            fpath_out=fpath_out,
            new_duration=new_duration,
            buff_size=self.buff_size,
        )
        return nevents

    def read(self, fpath: Optional[Union[str, pathlib.Path]] = None) -> ndarray:
        """
        Reads a binary file to a structured NumPy of events.

        :param fpath: path to the input file.

        :returns: the structured NumPy array.
        """
        fpath = check_external_file(fpath, self.fpath, self.encoding)
        arr, status = c_read_wrapper(
            encoding=self.encoding, fpath=fpath, buff_size=self.buff_size
        )
        if status != 0:
            raise RuntimeError(
                "ERROR: Something went wrong while creating the array from the file."
            )
        return arr

    def read_chunk(self) -> ndarray:
        """
        Generator used to read the file in chunks.

        :returns: structured NumPy array of events.
        """
        if self.fpath is None:
            raise ValueError("ERROR: An input file must be set.")
        while self.cargo.events_info.dim > 0:
            arr, self.cargo, status = c_read_chunk_wrapper(
                encoding=self.encoding,
                fpath=self.fpath,
                cargo=self.cargo,
                buff_size=self.buff_size,
            )
            if arr is None or status != 0:
                break
            yield arr



from pathlib import Path
import numpy as np
from typing import Union

_ROOT_PATH = Path(__file__).resolve().parent.parent

_SUPPORTED_ENCODINGS = (
    "DAT",
    "EVT2",
    "EVT3",
)

_DEFAULT_DTYPE = np.dtype(
    [("t", np.int64), ("x", np.int16), ("y", np.int16), ("p", np.uint8)]
)
_DEFAULT_BUFF_SIZE = 4096


def check_file_encoding(fpath: Union[str, Path], encoding: str) -> None:
    if encoding == "DAT":
        if not (str(fpath).endswith(".dat")):
            raise ValueError(
                "ERROR: The DAT encoding needs a '.dat' file to be specified."
            )
    elif encoding == "EVT2" or encoding == "EVT3":
        if not (str(fpath).endswith(".raw")):
            raise ValueError(
                "ERROR: The EVT2/EVT3 encoding needs a '.raw' file to be specified."
            )
    return


def check_encoding(encoding: str) -> str:
    if not (isinstance(encoding, str)):
        raise TypeError("ERROR: The encoding must be specified as a string.")
    encoding = encoding.upper()
    if not (encoding in _SUPPORTED_ENCODINGS):
        raise ValueError("ERROR: The encoding provided is not supported.")
    return encoding


def check_input_file(fpath: Union[str, Path], encoding: str) -> Union[str, Path]:
    if not (isinstance(fpath, str) or isinstance(fpath, Path)):
        raise TypeError(
            f"ERROR: The file path has to be provided as a string or a pathlib.Path object."
        )
    fpath = Path(fpath).resolve()
    if not (fpath.is_file()):
        raise ValueError("ERROR: The file provided does not exist.")
    check_file_encoding(fpath, encoding)
    return fpath


def check_output_file(fpath: Union[str, Path], encoding: str) -> Union[str, Path]:
    if not (isinstance(fpath, str) or isinstance(fpath, Path)):
        raise TypeError(
            "ERROR: The file path provided must be a string or a pathlib.Path object."
        )
    fpath = Path(fpath).resolve()
    if not fpath.parent.is_dir():
        raise NotADirectoryError("ERROR: The output file directory does not exist.")
    check_file_encoding(fpath, encoding)
    fpath.touch()
    if not fpath.is_file():
        raise SystemError("ERROR: The output file could not be created.")
    check_file_encoding(fpath, encoding)
    return fpath


def check_chunk_size(chunk_size: int, encoding: str) -> int:
    if not (isinstance(chunk_size, int)):
        raise TypeError("ERROR: The chunk size must be a positive integer number.")
    if encoding == "EVT3":
        if chunk_size < 12:
            raise ValueError(
                "ERROR: The chunk size has to be larger than 12 for the EVT3 encoding."
            )
    else:
        if chunk_size <= 0:
            raise ValueError("ERROR: The chunk size has to be larger than 0.")
    return chunk_size


def check_buff_size(buff_size: int) -> int:
    if not isinstance(buff_size, int):
        raise TypeError("ERROR: The buffer size must be a positive integer.")
    if buff_size <= 0:
        raise ValueError("ERROR: The buffer size must be larger than 0.")
    return buff_size


def check_new_duration(new_duration: int) -> int:
    if not (isinstance(new_duration, int)):
        raise TypeError(
            "ERROR: The new duration must be an integer positive number (not integer)."
        )
    if new_duration <= 0:
        raise ValueError("ERROR: The new duration must be larger than 0.")
    return new_duration


def check_external_file(
    fpath: Union[str, Path], self_fpath: Union[str, Path], encoding: str
) -> Union[str, Path]:
    if fpath is None:
        if self_fpath is None:
            raise ValueError("ERROR: An input file must be set or provided.")
        return self_fpath
    else:
        return check_input_file(fpath=fpath, encoding=encoding)



from pathlib import Path
from ctypes import c_char_p, c_size_t, byref
from typing import Optional, Union
from numpy import empty
from expelliarmus.utils import _SUPPORTED_ENCODINGS
from expelliarmus.wizard.clib import (
    event_t,
    events_cargo_t,
    dat_cargo_t,
    evt2_cargo_t,
    evt3_cargo_t,
    c_read_dat,
    c_read_evt2,
    c_read_evt3,
    c_measure_dat,
    c_measure_evt2,
    c_measure_evt3,
    c_cut_dat,
    c_cut_evt2,
    c_cut_evt3,
)


def c_read_wrapper(encoding: str, fpath: Union[str, Path], buff_size: int):
    # Error handling.
    if not isinstance(encoding, str):
        raise TypeError(
            "ERROR: The encoding must be a string among 'DAT', 'EVT2' and 'EVT3'."
        )
    if not (isinstance(fpath, str) or isinstance(fpath, Path)):
        raise TypeError(
            "ERROR: The file path must be a string or a pathlib.Path object."
        )
    if not isinstance(buff_size, int):
        raise TypeError("ERROR: The buffer size must be an integer larger than 0.")
    if encoding not in _SUPPORTED_ENCODINGS:
        raise ValueError("ERROR: The encoding provided is not valid.")
    if buff_size <= 0:
        raise ValueError("ERROR: The buffer size must be larger than 0.")

    c_fpath = c_char_p(bytes(str(fpath), "utf-8"))
    c_buff_size = c_size_t(buff_size)
    if encoding == "DAT":
        dim = c_measure_dat(c_fpath, c_buff_size)
        events_info = events_cargo_t(dim, 0, 0)
        cargo = dat_cargo_t(events_info, 0, 0)
    elif encoding == "EVT2":
        dim = c_measure_evt2(c_fpath, c_buff_size)
        events_info = events_cargo_t(dim, 0, 0)
        cargo = evt2_cargo_t(events_info, 0, 0)
    elif encoding == "EVT3":
        dim = c_measure_evt3(c_fpath, c_buff_size)
        events_info = events_cargo_t(dim, 0, 0)
        last_event = event_t(0, 0, 0, 0)
        cargo = evt3_cargo_t(events_info, 0, 0, 0, 0, 0, last_event)
    if dim > 0:
        arr = empty((dim,), dtype=event_t)
        if encoding == "DAT":
            status = c_read_dat(c_fpath, arr, byref(cargo), c_buff_size)
        elif encoding == "EVT2":
            status = c_read_evt2(c_fpath, arr, byref(cargo), c_buff_size)
        elif encoding == "EVT3":
            status = c_read_evt3(c_fpath, arr, byref(cargo), c_buff_size)
    return (arr, status) if dim > 0 and status == 0 else (None, status)


def c_read_chunk_wrapper(
    encoding: str,
    fpath: Union[str, Path],
    cargo: Union[dat_cargo_t, evt2_cargo_t, evt3_cargo_t],
    buff_size: int,
):
    if not (isinstance(fpath, str) or isinstance(fpath, Path)):
        raise TypeError(
            "ERROR: The file path must be a string or a pathlib.Path object."
        )
    if not isinstance(buff_size, int):
        raise TypeError("ERROR: The buffer size must be an integer larger than 0.")
    if encoding not in _SUPPORTED_ENCODINGS:
        raise ValueError("ERROR: The encoding provided is not valid.")
    if buff_size <= 0:
        raise ValueError("ERROR: The buffer size must be larger than 0.")
    if not (
        isinstance(cargo, dat_cargo_t)
        or isinstance(cargo, evt2_cargo_t)
        or isinstance(cargo, evt3_cargo_t)
    ):
        raise TypeError("ERROR: The cargo type is not supported.")

    c_fpath = c_char_p(bytes(str(fpath), "utf-8"))
    c_buff_size = c_size_t(buff_size)
    if encoding == "DAT":
        arr = empty((cargo.events_info.dim,), dtype=event_t)
        status = c_read_dat(c_fpath, arr, byref(cargo), c_buff_size)
    elif encoding == "EVT2":
        arr = empty((cargo.events_info.dim,), dtype=event_t)
        status = c_read_evt2(c_fpath, arr, byref(cargo), c_buff_size)
    elif encoding == "EVT3":
        arr = empty((cargo.events_info.dim + 12,), dtype=event_t)
        status = c_read_evt3(c_fpath, arr, byref(cargo), c_buff_size)
    return (
        (arr[: cargo.events_info.dim], cargo, status)
        if status == 0
        else (None, cargo, status)
    )


def c_cut_wrapper(
    encoding: str,
    fpath_in: Union[str, Path],
    fpath_out: Union[str, Path],
    new_duration: int,
    buff_size: int,
):
    if not (isinstance(fpath_in, str) or isinstance(fpath_in, Path)):
        raise TypeError(
            "ERROR: The input file path must be a string or a pathlib.Path object."
        )
    if not (isinstance(fpath_out, str) or isinstance(fpath_out, Path)):
        raise TypeError(
            "ERROR: The output file path must be a string or a pathlib.Path object."
        )
    if not isinstance(buff_size, int):
        raise TypeError("ERROR: The buffer size must be an integer larger than 0.")
    if buff_size <= 0:
        raise ValueError("ERROR: The buffer size must be larger than 0.")
    if encoding not in _SUPPORTED_ENCODINGS:
        raise ValueError("ERROR: The encoding provided is not valid.")
    if not isinstance(new_duration, int):
        raise TypeError("ERROR: The new duration must be a positive integer.")
    if new_duration <= 0:
        raise ValueError("ERROR: The new duration must be larger than zero.")

    c_fpath_in = c_char_p(bytes(str(fpath_in), "utf-8"))
    c_fpath_out = c_char_p(bytes(str(fpath_out), "utf-8"))
    c_new_duration = c_size_t(new_duration)
    c_buff_size = c_size_t(buff_size)
    if encoding == "DAT":
        c_dim = c_cut_dat(c_fpath_in, c_fpath_out, c_new_duration, c_buff_size)
    elif encoding == "EVT2":
        c_dim = c_cut_evt2(c_fpath_in, c_fpath_out, c_new_duration, c_buff_size)
    elif encoding == "EVT3":
        c_dim = c_cut_evt3(c_fpath_in, c_fpath_out, c_new_duration, c_buff_size)
    return c_dim



from ctypes import (
    CDLL,
    c_char_p,
    POINTER,
    c_size_t,
    c_int64,
    c_int16,
    c_uint16,
    c_uint8,
    c_uint,
    c_int,
    c_uint64,
    Structure,
)
import pathlib
import os
import re
from expelliarmus.utils import _ROOT_PATH
from numpy.ctypeslib import ndpointer
from numpy import zeros

# Searching for the shared library.
lib_re = r"^expelliarmus\..*\.(so|pyd)$"
for root, dirs, files in os.walk(_ROOT_PATH):
    for f in files:
        if re.match(lib_re, f):
            lib_path = pathlib.Path(os.path.join(root, f))
            break
clib = CDLL(str(lib_path))


class event_t(Structure):
    _fields_ = [
        ("t", c_int64),
        ("x", c_int16),
        ("y", c_int16),
        ("p", c_uint8),
    ]


class events_cargo_t(Structure):
    _fields_ = [
        ("dim", c_size_t),
        ("is_chunk", c_uint8),
        ("bytes_read", c_size_t),
    ]


class dat_cargo_t(Structure):
    _fields_ = [
        ("events_info", events_cargo_t),
        ("last_t", c_int64),
        ("time_ovfs", c_uint64),
    ]


class evt2_cargo_t(Structure):
    _fields_ = [
        ("events_info", events_cargo_t),
        ("last_t", c_int64),
        ("time_high", c_uint64),
    ]


class evt3_cargo_t(Structure):
    _fields_ = [
        ("events_info", events_cargo_t),
        ("time_high", c_uint64),
        ("time_low", c_uint64),
        ("time_high_ovfs", c_uint64),
        ("time_low_ovfs", c_uint64),
        ("base_x", c_uint16),
        ("last_event", event_t),
    ]


# Setting up C wrappers.
# Read functions.
c_read_dat = clib.read_dat
c_read_dat.argtypes = [
    c_char_p,
    ndpointer(dtype=event_t, ndim=1),
    POINTER(dat_cargo_t),
    c_size_t,
]
c_read_dat.restype = c_int

c_read_evt2 = clib.read_evt2
c_read_evt2.argtypes = [
    c_char_p,
    ndpointer(dtype=event_t, ndim=1),
    POINTER(evt2_cargo_t),
    c_size_t,
]
c_read_evt2.restype = c_int

c_read_evt3 = clib.read_evt3
c_read_evt3.argtypes = [
    c_char_p,
    ndpointer(dtype=event_t, ndim=1),
    POINTER(evt3_cargo_t),
    c_size_t,
]
c_read_evt3.restype = c_int

# Cut functions.
ARGTYPES_CUT = [c_char_p, c_char_p, c_size_t, c_size_t]
RESTYPE_CUT = c_size_t

c_cut_dat = clib.cut_dat
c_cut_evt2 = clib.cut_evt2
c_cut_evt3 = clib.cut_evt3
for c_cut_fn in (c_cut_dat, c_cut_evt2, c_cut_evt3):
    c_cut_fn.restype = RESTYPE_CUT
    c_cut_fn.argtypes = ARGTYPES_CUT

# Measure functions.
ARGTYPES_MEASURE = [c_char_p, c_size_t]
RESTYPE_MEASURE = c_size_t

c_measure_dat = clib.measure_dat
c_measure_evt2 = clib.measure_evt2
c_measure_evt3 = clib.measure_evt3
for fn in (c_measure_dat, c_measure_evt2, c_measure_evt3):
    fn.argtypes = ARGTYPES_MEASURE
    fn.restype = RESTYPE_MEASURE



# Inspired by https://github.com/himbeles/ctypes-example.
import pathlib
from distutils.command.build_ext import build_ext as build_ext_orig

from setuptools import Extension, setup


class CTypesExtension(Extension):
    pass


class build_ext(build_ext_orig):
    def build_extension(self, ext):
        self._ctypes = isinstance(ext, CTypesExtension)
        return super().build_extension(ext)

    def get_export_symbols(self, ext):
        if self._ctypes:
            return ext.export_symbols
        return super().get_export_symbols(ext)

    def get_ext_filename(self, ext_name):
        return super().get_ext_filename(ext_name)


with open("README.md", "r") as file:
    long_description = file.read()

setup(
    name="expelliarmus",
    install_requires=["numpy"],
    author="Fabrizio Ottati, Gregor Lenz",
    author_email="fabriziottati@gmail.com, mail@lenzgregor.com",
    maintainer="Fabrizio Ottati, Gregor Lenz",
    maintainer_email="fabriziottati@gmail.com, mail@lenzgregor.com",
    url="https://github.com/fabhertz95/expelliarmus",
    version="1.1.3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["expelliarmus", "expelliarmus.wizard"],
    ext_modules=[
        CTypesExtension(
            "expelliarmus",
            [
                str(pathlib.Path("expelliarmus", "src", "wizard.c")),
            ],
        ),
    ],
    cmdclass={"build_ext": build_ext},
)
